(* +------------------------------------------------------------------------+
   |                                                                        |
   |                                                                        |
   |                 Boite  outils pour les manipulations d'arbres         |
   |                                                                        |
   |                                                                        |
   +------------------------------------------------------------------------+ *)
(* Michel Quercia, le 10/01/98 *)

                     (* +-------------------------------+
                        |  Fonctions de transformation  |
                        +-------------------------------+ *)

let rec g_map f = function
  G_noeud(x,fils) -> G_noeud(f(x), map (g_map f) fils);;

let rec b_map f = function
| B_vide -> B_vide
| B_noeud(x,g,d) -> B_noeud(f(x), b_map f g, b_map f d);;


                     (* +------------------------------+
                        |  Fonctions de mise en arbre  |
                        +------------------------------+ *)


type terme = Pg | Pd | Eos | Terme of string;;

let rec mot_suivant(ch,i) =
  let l = string_length(ch) in
  if i >= l then (Eos,i)
  else match ch.[i] with
  | (` `|`\t`|`\n`) -> mot_suivant(ch,i+1)
  | `(` -> (Pg,i+1)
  | `)` -> (Pd,i+1)
  | _ -> let j = ref(i+1) and k = ref(i+1) and encore = ref(true) in
         while !j < l & !encore do match ch.[!j] with
         | (` `|`\t`|`\n`) -> incr j
	 | (`(`| `)`)      -> encore := false
         | _               -> incr j; k := !j
         done;
         (Terme(sub_string ch i (!k - i)), !j)
;; 

let erreur msg i = raise(Erreur((Failure msg),i));;

let rec g_analyse(ch,i) = match mot_suivant(ch,i) with
| (Pg,j) -> begin
              match mot_suivant(ch,j) with
              | (Terme(e),k) -> let (f,l) = f_analyse(ch,k) in (G_noeud(e,f),l)
              | _ -> erreur "il manque un terme" j
            end
| _ -> erreur "il manque \"(\"" i

and f_analyse(ch,i) = match mot_suivant(ch,i) with
| (Pd,j) -> ([],j)
| _ -> let (a,j) = g_analyse(ch,i) in
       let (f,k) = f_analyse(ch,j) in (a::f,k)

and b_analyse(ch,i) = match mot_suivant(ch,i) with
  | (Pg,j) ->
    begin
      match mot_suivant(ch,j) with
      | (Terme(e),k) ->
        begin
          let (g,l) = b_analyse(ch,k) in
          let (d,m) = b_analyse(ch,l) in
          match mot_suivant(ch,m) with
          | (Pd,n) -> (B_noeud(e,g,d),n)
          | _ -> erreur "il manque \")\"" m
        end
      | (Pd,k) -> (B_vide,k)
      | _ -> erreur "il manque un terme" j
    end
  | _ -> (B_vide,i)
;;

let string_x_arbre analyseur ch =
  let (a,i) = analyseur(ch,0) in
  match mot_suivant(ch,i) with
  | (Eos,_) -> a
  | _ -> erreur "expression hors parenthses" i
;;

let string_g_arbre(ch) = string_x_arbre g_analyse ch
and string_b_arbre(ch) = string_x_arbre b_analyse ch;;

let int_g_arbre(ch) = g_map int_of_string (string_g_arbre ch)
and int_b_arbre(ch) = b_map int_of_string (string_b_arbre ch);;


        (* +----------------------------------------------+
           |  Conversion arbre -> expression parenthse  |
           +----------------------------------------------+ *)

(* copie [ch] dans [!s.[!i...]] et avance [i] en consquence *)
(* gre les cas de dbordement *)
let append s i ch =
  let l = string_length(!s)  in
  let l' = string_length(ch) in
  if l'+ !i > l then begin
    let s' = create_string (2*l+l') in
    blit_string !s 0 s' 0 !i;
    s := s'
  end;
  blit_string ch 0 !s !i l';
  i := !i+l'
;;
  
let rec s_string_g_arbre s i (G_noeud(tiq,fils)) =
  append s i " (";
  append s i tiq;
  do_list (s_string_g_arbre s i) fils;
  append s i ")"
;;

let rec s_string_b_arbre s i = function
| B_vide -> append s i " ()"
| B_noeud(tiq,g,d) ->
  append s i " (";
  append s i tiq;
  s_string_b_arbre s i g;
  s_string_b_arbre s i d;
  append s i ")"
;;

let string_of_string_x_arbre conv a =
  let s = ref "" and i = ref 0 in
  conv s i a;
  sub_string !s 1 (!i-1)
;;

let string_of_string_g_arbre(a) = string_of_string_x_arbre s_string_g_arbre a
and string_of_string_b_arbre(a) = string_of_string_x_arbre s_string_b_arbre a
;;

let string_of_int_g_arbre(a) = string_of_string_g_arbre(g_map string_of_int a)
and string_of_int_b_arbre(a) = string_of_string_b_arbre(b_map string_of_int a)
;;


    (* +------------------------------------------------------+
       |  Imprime un arbre gnral sous forme semi-graphique  |
       +------------------------------------------------------+ *)

#open "format";;

let centre chaine l =
  let lc = string_length(chaine) in
  if lc >= l then chaine
             else let g = (l-lc)/2 in let d = l-lc-g in
                  (make_string g `_`) ^ chaine ^(make_string d `_`)

;;

(* calcule les abscisses relatives d'impression  *)
(* des noeuds et l'encombrement total d'un arbre *)
let rec abscisses_g_arbre x0 = function G_noeud(tiq,fils) ->

  let (enc,xf0,xfn,fort) = abscisses_fort 0 fils in
  let xf = (xf0+xfn)/2 in
  let fort' = g_dcale (-xf) fort in

  let tiq' = centre tiq (xfn-xf0-3) in
  let l = string_length(tiq') in
  (max enc l, G_noeud((x0+(max xf (l/2)),tiq'),fort'))

and abscisses_fort x0 = function
| []  -> (0,0,0,[])
| [a] -> let (enc,(G_noeud((x,_),_) as b)) = abscisses_g_arbre x0 a in (enc,x,x,[b])
| a::suite ->
    let (enc,(G_noeud((x,_),_) as b)) = abscisses_g_arbre x0 a in
    let (enc',_,y,f) = abscisses_fort (x0+enc+2) suite in (enc+enc'+2,x,y,b::f)


and g_dcale dx = function
| [] -> []
| G_noeud((x,e),f)::suite -> G_noeud((x+dx,e),f) :: g_dcale dx suite

;;


let rec print_g_noeuds x0 = function
| [] -> if x0 > 0 then print_cut()
| G_noeud((x,e),_)::suite ->
    let l = string_length(e) in
    for i=x0 to x-(l-1)/2-1 do print_char ` ` done;
    print_string(e);
    print_g_noeuds (x+l/2+1) suite
;;

let rec print_g_traits_fils x0 x = function
| [] -> x0
| G_noeud((y,_),_)::suite ->
    let z = (if y < 0 then x+y else if y = 0 then x+y-1 else x+y-2) in
    for i=x0 to z do print_char ` ` done;
    print_char (if y < 0 then `/` else if y = 0 then `|` else `\\`);
    print_g_traits_fils (z+2) x suite
;;

let rec print_g_traits x0 = function
| [] -> if x0 > 0 then print_cut()
| G_noeud((x,_),f)::suite ->
    let x1 = print_g_traits_fils x0 x f in
    print_g_traits x1 suite
;;

let rec liste_g_fils = function
| [] -> []
| G_noeud((x,e),f)::suite -> (g_dcale x f) @ (liste_g_fils suite)
;;

let rec print_g_fort = function
| [] -> ()
| f  -> print_g_noeuds 0 f;
        print_g_traits 0 f;
        print_g_fort(liste_g_fils f)
;;

let print_string_g_arbre a =
  open_vbox(0);
  print_g_fort [snd(abscisses_g_arbre 0 a)];
  close_box()
;;

let print_int_g_arbre(a) = print_string_g_arbre(g_map string_of_int a);;

    (* +------------------------------------------------------+
       |  Imprime un arbre binaire sous forme semi-graphique  |
       +------------------------------------------------------+ *)

(* calcule les abscisses relatives d'impression  *)
(* des noeuds et l'encombrement total d'un arbre *)
let rec abscisses_b_arbre = function

| B_vide -> (0,B_vide)

| B_noeud(tiq,B_vide,B_vide) ->
   let l = string_length(tiq) in (l, B_noeud((l/2,tiq),B_vide,B_vide))

| B_noeud(tiq,g,B_vide) ->
   let l = string_length(tiq)
   and (eg,xg,g') = match abscisses_b_arbre g with
   | (eg,B_noeud((x,e),g,d)) -> (eg,x,B_noeud((-1,e),g,d))
   | _ -> failwith "cas impossible"
   in
   let x0 = min (-l/2) (-xg-1) and x1 = max (l-l/2) (eg-xg-1)
   in (x1-x0, B_noeud((-x0,tiq),g',B_vide))

| B_noeud(tiq,B_vide,d) ->
   let l = string_length(tiq)
   and (ed,xd,d') = match abscisses_b_arbre d with
   | (ed,B_noeud((x,e),g,d)) -> (ed,x,B_noeud((1,e),g,d))
   | _ -> failwith "cas impossible"
   in
   let x0 = min (-l/2) (-xd+1) and x1 = max (l-l/2) (ed-xd+1)
   in (x1-x0, B_noeud((-x0,tiq),B_vide,d'))

| B_noeud(tiq,g,d) ->

   let (eg,xg,tig,gg,gd) = match abscisses_b_arbre g with
   | (eg,B_noeud((x,tig),gg,gd)) -> (eg,x,tig,gg,gd)
   | _ -> failwith "cas impossible"
   in
  
   let (ed,xd,tid,dg,dd) = match abscisses_b_arbre d with
   | (ed,B_noeud((x,tid),dg,dd)) -> (ed,eg+x+2,tid,dg,dd)
   | _ -> failwith "cas impossible"
   in

   let xm = (xd+xg)/2 in
   let tiq' = centre tiq (xd-xg-3) in
   let l = string_length(tiq')
   in (max (eg+ed+2) l, B_noeud((max xm (l/2),tiq'),
                          B_noeud((xg-xm,tig),gg,gd),
                          B_noeud((xd-xm,tid),dg,dd)))

;;


(* impression *)
let rec print_b_noeuds x0 = function
| [] -> if x0 > 0 then print_cut()
| B_vide::suite -> print_b_noeuds x0 suite
| B_noeud((x,e),_,_)::suite ->
    let l = string_length(e) in
    for i=x0 to x-(l-1)/2-1 do print_char ` ` done;
    print_string(e);
    print_b_noeuds (x+l/2+1) suite
;;

let print_b_traits_fils x0 x = function
| B_vide -> x0
| B_noeud((y,_),_,_) ->
    let z = (if y < 0 then x+y else if y = 0 then x+y-1 else x+y-2) in
    for i=x0 to z do print_char ` ` done;
    print_char (if y < 0 then `/` else if y = 0 then `|` else `\\`);
    z+2
;;

let rec print_b_traits x0 = function
| [] -> if x0 > 0 then print_cut()
| B_vide::suite -> print_b_traits x0 suite
| B_noeud((x,_),g,d)::suite ->
    let x1 = print_b_traits_fils x0 x g in
    let x2 = print_b_traits_fils x1 x d in
    print_b_traits x2 suite
;;

let rec liste_b_fils = function
| [] -> []
| B_vide::suite -> liste_b_fils suite
| B_noeud((x,e),g,d)::suite ->
   (match g with B_vide -> [] | B_noeud((y,e),g,d) -> [B_noeud((x+y,e),g,d)])
 @ (match d with B_vide -> [] | B_noeud((y,e),g,d) -> [B_noeud((x+y,e),g,d)])
 @ (liste_b_fils suite)
;;

let rec print_b_fort = function
| [] -> ()
| f  -> print_b_noeuds 0 f;
        print_b_traits 0 f;
        print_b_fort(liste_b_fils f)
;;

let print_string_b_arbre a =
  open_vbox(0);
  print_b_fort [snd(abscisses_b_arbre a)];
  close_box()
;;

let print_int_b_arbre(a) = print_string_b_arbre(b_map string_of_int a);;

(* +------------------------------------------------------------------------+
   |                                                                        |
   |                       Evaluation de formules infixes                   |
   |                                                                        |
   +------------------------------------------------------------------------+ *)

(*
 * [expr impl p flux] value le dbut de [flux] jusqu' rencontrer un
 *   oprateur infixe de priorit infrieure ou gale  [p].
 *   [impl] retourne un couple (priorit,opration infixe implicite)
 *   ou dclenche [Parse_failure] s'il n'y a pas d'opration implicite.
 *
 * [atome flux] envoie la premire valeur dans [flux] en appliquant
 *   les oprateurs prfixe et les regroupement entre parenthses.
 *
 * [liste flux] et [fin_liste flux] assemblent une liste de valeurs.
 *
 * [suite_expr p a flux] applique les oprateurs postfixe et les oprateurs
 *   infixe de priorit suprieure  [p]  la valeur [a].
 *
 * [infixe_prioritaire p flux] renvoie un'oprateur infixe s'il figure
 *   en tte de [flux] et a une priorit suprieure  [p], choue sinin.
 *)

let rec expr impl p = function [< (atome impl) a; (suite_expr impl p a) b >] -> b

and atome impl = function
| [< 'Valeur v >] -> v
| [< 'Prfixe f; (atome impl) v >] -> f v
| [< 'InfPref(_,f); (atome impl) v >] -> f v
| [< 'Lpar; (expr impl 0) v; 'Rpar >] -> v
| [< 'Liste f; 'Lpar; (liste impl) l; 'Rpar >] -> f l

and liste impl = function
| [< (expr impl 0) v; (fin_liste impl) l >] -> v :: l
| [< >] -> []

and fin_liste impl = function
| [< 'Lsep; (expr impl 0) v; (fin_liste impl) l >] -> v :: l
| [< >] -> []

and suite_expr impl p a = function
| [< 'Postfixe f; (suite_expr impl p (f a)) b >] -> b
| [< (infixe_prioritaire impl p) (q,f); (expr impl q) b; (suite_expr impl p (f a b)) c >] -> c
| [< >] -> a

and infixe_prioritaire impl p flux = match stream_get flux with
| Infixe(q,f),_      -> if q > p then (stream_next flux; (q,f)) else raise Parse_failure
| InfPref((q,f),_),_ -> if q > p then (stream_next flux; (q,f)) else raise Parse_failure
| Rpar,_ -> raise Parse_failure
| Lsep,_ -> raise Parse_failure
| _ -> impl()
;; 

(*
 * [analyse impl f lexer s] analyse la chaine [s] et renvoie la valeur qu'elle
 * reprsente. [s] est d'abord convertie en flux de caractres, ce flux est
 * soumis  l'analyseur lexical [lexer], puis les lexmes obtenus sont
 * convertis en lments syntaxiques par [f]. [impl()] retourne un couple
 * (priorit, opration infixe implicite) ou dclenche [Parse_failure]
 *)

let analyseur impl syntax_of_lexme lexer s =

  (*
   * fouttoir avec le moins : s'il est immdiatement suivi d'un entier,
   * genlex produit une constante ngative et l'oprateur "-" n'apparat
   * pas dans le flux. Solution : on ajoute un espace fictif entre un "-"
   * et un chiffre.
   *)
  let i     = ref (-1)      (* index chaine       *)
  and moins = ref false     (* c. prcdent = "-" *)
  and lg    = string_length s in

  let ch() =
    if !moins & !i+1 < lg & s.[!i+1] >= `0` & s.[!i+1] <= `9`
    then begin
       moins := false;
       ` `
    end
    else begin
      incr i;
      if !i >= lg then raise Parse_failure
      else begin
        moins := s.[!i] = `-`;
        s.[!i]
      end
    end

  in

  let st = lexer(stream_from ch) in
  let stx() = syntax_of_lexme(stream_next st) in
  try match stream_from stx with [< (expr impl 0) v; end_of_stream _ >] -> v
  with ex -> raise(Erreur(ex,!i))
;;

(* impression d'un chaine erronne *)
let print_erreur s i =
  printf__fprintf stderr "%s\n" s;
  printf__fprintf stderr "%s^\n" (make_string i ` `);
  flush stderr
;;
