             (* +------------------------------------+
                |  Exemple d'utilisation de graphes  |
                +------------------------------------+ *)

open Graphes;;
open Graphx;;
#load "mappe.cmo";;
#load "graphes.cmo";;

(* conversion formule -> DAG *)
let green = rgb 0 192 0;;
let graphe(formule) =

  let g = ref(new_graphe()) in

  let rec job(f) = match f with
  | Const(c)     -> g := add_noeud !g f (string_of_float c)
  | Var(x)       -> g := add_noeud !g f x
  | Fct(nom,u)   -> job(u);
                    g := add_noeud !g f nom;
                    g := add_arc   !g (u,f) ""
  | Bin(op,u,v)  -> job(u); job(v);
                    g := add_noeud !g f op;
                    g := add_arc   !g (u,f) "";
                    g := add_arc   !g (v,f) ""
  in job(formule);
  !g
;;

(* calcul des positions *)
let xgraphe g formule =

  let couleur = function
  | (Const(_) | Var(_)) -> green
  | _ -> blue
  and rayon(f) = if f = formule then -20 else 20
  in

  make_xgraphe
    (fun (f,tiquette) -> (couleur f, rayon f, tiquette)) (* noeuds *)
    (fun _ -> (red,10,""))                                 (* arcs   *)
    100                                          (* pas de la grille *)
    g                                                      (* graphe *)
    formule                                      (* noeud de dpart  *)
;;

(* affichage *)
let show(xg) = edit 25 xg "" [];;


