(* +------------------------------------------------------------------------+
   |                                                                        |
   |                                                                        |
   |                Manipulation et visualisation de graphes                |
   |                                                                        |
   |                                                                        |
   +------------------------------------------------------------------------+ *)
(* M. Quercia, le 13/04/98 *)

(*
  structure asbtraite : 'a est le type des noeuds, 'b le type des
  tiquettes des noeuds et 'c le type des tiquettes des arcs.
*)
type ('a,'b,'c) graphe;;

                  (* +---------------------------+
                     |  Construction de graphes  |
                     +---------------------------+ *)

(* le graphe fourni en argument n'est pas modifi, un nouveau graphe est retourn. *)

val new_graphe : unit -> ('a,'b,'c) graphe
val add_noeud  : ('a,'b,'c) graphe -> 'a -> 'b      -> ('a,'b,'c) graphe
val del_noeud  : ('a,'b,'c) graphe -> 'a            -> ('a,'b,'c) graphe
val add_arc    : ('a,'b,'c) graphe -> ('a*'a) -> 'c -> ('a,'b,'c) graphe
val del_arc    : ('a,'b,'c) graphe -> ('a*'a)       -> ('a,'b,'c) graphe
;;
(*
 [new_graphe()]      retourne un graphe sans noeuds ni arcs.
 [add_noeud g n e]   ajoute le noeud [n] d'tiquette [e]. Remplace l'tiquette
                     de [n] si [n] est dj prsent dans [g].
 [del_noeud g n]     supprime le noeud [n] son tiquette et tous les arcs associs
                     si [n] figure dans [g], retourne [g] inchang sinon.
 [add_arc g (n,p) e] ajoute un arc de [n] vers [p] d'tiquette [e]. Remplace
                     l'tiquette de [n->p] s'il y a un tel arc dans [g].
                     Dclenche l'exception [Not_found] si [n] ou [p] ne
                     figurent pas dans [g].
 [del_arc g (n,p)]   suprime l'arc de [n] vers [p]. Retourne [g] s'il n'y
                     a pas de tel arc dans [g].
*)

val mirroir : ('a,'b,'c) graphe -> ('a,'b,'c) graphe;;
(* Retourne les flches. *)

                 (* +----------------------------+
                    |  Consultation d'un graphe  |
                    +----------------------------+ *)

val noeuds  : ('a,'b,'c) graphe -> 'a list
val arcs    : ('a,'b,'c) graphe -> (('a*'a)*'c) list
val member  : ('a,'b,'c) graphe -> 'a -> bool
val succ    : ('a,'b,'c) graphe -> 'a -> 'a list
val pred    : ('a,'b,'c) graphe -> 'a -> 'a list
val etiq    : ('a,'b,'c) graphe -> 'a -> 'b
val arc     : ('a,'b,'c) graphe -> ('a*'a) -> 'c
val arc_opt : ('a,'b,'c) graphe -> ('a*'a) -> 'c option
;;
(*
 [noeuds g]    retourne la liste des noeuds de [g].
 [arcs g]      retourne la liste des arc de [g]. Un arc de [n] vers [p]
               d'tiquette [e] ets cod par [((n,p),e)].
 [member g n]  dit si [n] figure dans [g].
 [succ g n]    retourne la liste des successeurs de [n] dans [g]. Retourne une
               liste vide si [n] ne figure pas dans [g].
 [pred g n]    retourne la liste des prdecesseurs de [n] dans [g]. Retourne une
               liste vide si [n] ne figure pas dans [g].
 [etiq g n]    retourne l'tiquette du noeud [n]. Dclenche l'exception
               [Not_found] si [n] ne figure pas dans [g].
 [arc g (n,p)] retourne l'tiquette de l'arc de [n] vers [p] dans [g].
               Dclenche l'exception [Not_found] s'il n'y a pas de tel arc.
 [arc_opt g (n,p)] retourne [Some(e)] s'il y a dans [g] un arc de [n] vers
               [p] d'tiquette [e] et [None] sinon.
*)

                   (* +------------------------+
                      |  Parcours d'un graphe  |
                      +------------------------+ *)


val do_noeuds : ('a*'b -> unit) -> ('a,'b,'c) graphe -> unit
val do_arcs   : (('a*'a)*'c -> unit) -> ('a,'b,'c) graphe -> unit
;;
(*
 [do_noeuds f g] applique [f]  chaque noeud de [g] dans un ordre non spcifi.
 [do_arcs f g]   applique [f]  chaque arc de [g] dans un ordre non spcifi.
*)

                              (* +-------------+
                                 |  Affichage  |
                                 +-------------+ *)


open Graphx;;

type xnoeud = {
 npos   : point;     (* position d'affichage *)
 ncoul  : color;     (* couleur d'affichage  *)
 radius : int;       (* rayon du cercle (pas de cercle si =0, cercle double si <0) *)
 nlpos  : point;     (* position de la lgende *)
 nltxt  : string     (* texte de la lgende, pas de lgende si "" *)
}
and xarc = {
 apos  : point array; (* points de Bzier dfinissant le trac *)
 acoul : color;       (* couleur d'affichage *)
 arrow : int;         (* taille pointe de la flche (pas de pointe si <=0) *)
 alpos : point;       (* position de la lgende *)
 altxt : string       (* texte de la lgende, pas de lgende si "" *)
};;

type 'a xgraphe = ('a,xnoeud,xarc) graphe;;
val make_xgraphe : ('a*'b -> color*int*string) ->
                   (('a*'a)*'c -> color*int*string) ->
                   int -> 
                   ('a,'b,'c) graphe -> 
                   'a -> 'a xgraphe;;
(*
 [make_xgraphe f g pas gr n] calcule les positions des noeuds et
    des arcs du graphe [gr]. [n] dsigne un noeud initial devant tre
    plac le plus  gauche, les autres noeuds sont placs en fonction de
    leur distance en arcs  [n] sur un rseau de taille [pas].
    Les noeuds n'appartenant pas  la composante connexe de [n] sont
    placs arbitrairement.
    [f(n,e)] retourne les informations [(ncoul,radius,nltxt)] pour un noeud
    [n] d'tiquette [e] et [g ((n,p),e)] retourne les informations
    [(acoul,arrow,altxt)] pour un arc [n->p] d'tiquette [e].
*)

val draw_xnoeud  : color -> xnoeud -> unit
val draw_xarc    : int -> int -> xarc -> unit
val draw_ctl     : xarc -> unit
val draw_legende : point -> string -> unit
val draw_grille  : int -> unit
val display      : 'a xgraphe -> unit
;;
(*
 Fonctions d'affichage.
 Le paramtre [color] de [draw_xnoeud] dsigne la couleur de fond du disque
 Les paramtres [int] de [draw_xarc] dsignent les rayons des cercles
 entourant les extrmits de l'arc.
 [draw_ctl] trace le polygone des points de contrle d'un arc.
 [draw_grille pas] trace une grille de pas [pas].
 *)


                   (* +-------------------------+
                      |  Animation  la souris  |
                      +-------------------------+ *)

val clrscr : string -> unit;;
(* efface l'cran et affiche le texte spcifi en {x=0; y=0} *)

val select_noeud : 'a xgraphe -> (Graphx.status * 'a option)
val select_leg   : 'a xgraphe -> (Graphx.status * ('a option * ('a*'a) option))
val select_ctl   : 'a xgraphe -> (Graphx.status * ((('a*'a)*int) option));;
(*
 attend qu'un objet soit dsign  la souris ou qu'une touche soit enfonce.
 renvoie l'vnement dtect et l'objet ventuel dsign.

 select_noeud -> None ou Some(noeud)
 select_leg   -> (None ou Some(noeud), None ou Some(arc))
 select_ctl   -> Some(arc,numro du point)
 *)


val move_noeuds : int -> 'a xgraphe -> 'a xgraphe
val move_legs : 	 'a xgraphe -> 'a xgraphe
val move_ctls : 	 'a xgraphe -> 'a xgraphe;;
(*
 Dplace  la souris des noeuds, lgendes ou points de contrle jusqu'
 l'appui sur une touche; retourne le nouvel xgraphe.

 Le paramtre entier dans [move_noeuds] indique, s'il est positif, le
 pas d'un rseau servant  arrondir les positions des noeuds.

 La fonction [move_ctls] permet, aprs slection d'un point de contrle
 de modifier le nombre de segments de Bzier de l'arc dsign : en maintenant
 le bouton de la souris enfonc, appuyer sur '+' ou 'a' pour ajouter un
 segment, et sur '-' ou 's' pour supprimer un segment. La suppression est
 refuse s'il n'y a qu'un seul segment.
 *)

val postscript: 'a xgraphe -> unit;;
(* Enregistre le graphe dans un fichier au format postscript. *)

val get_string : string * string -> string;;
(*
 [get_string prompt init] saisit une chaine de caractre en bas de l'cran.
 [prompt] est le message d'invite, [init] la rponse par dfaut.
 Lve l'exception [Exit] si on appuie sur <Esc>.
 Les touches d'dition reconnues sont :

 BS  (\008) = effacement arrire
 DEL (\004 ou \127) = effacement avant
 ^B  (\002) = recule d'un caractre
 ^F  (\006) = avance d'un caractre
 ^K  (\011) = efface la fin de ligne
 ^A  (\001) = va en dbut de ligne
 ^E  (\005) = va en fin de ligne
 *)

val edit : int -> 'a xgraphe -> string ->
          (char * ('a xgraphe -> 'a xgraphe)) list ->
          'a xgraphe;;

(*
 [edit pas xg msg liste] affiche le graphe [xg] et traite les actions
 clavier jusqu' l'appui sur <Esc>.
 Les touches reconnues par dfaut sont :

   <n> : dplacement des noeuds
   <l> : dplacement des lgendes
   <c> : dplacement des points de contrle
   <p> : sortie postscript

 D'autres oprations peuvent tre implmentes en fournissant
 la liste des couples (touche,fonction) dans le paramtre [liste].
 Si l'une des touches spcifies est frappe, la fonction correspondante
 est appele avec le xgraphe courrant, elle doit retourner le nouvel xgraphe.
 Si les touches <n>,<l>,<c>,<p> figurent dans [liste] les fonctions
 correspondantes sont appeles au lieu de celles par dfaut.

 Le paramtre [msg] indique un texte  afficher en bas de l'cran
  titre de menu. S'il est vide, le menu par dfaut est affich.
 *)
